<?php

/**
 * JSON Web Token implementation, based on this spec:
 * https://tools.ietf.org/html/rfc7519
 *
 * PHP version 5
 *
 * @category Authentication
 * @package  Authentication_JWT
 * @author   Neuman Vong <neuman@twilio.com>
 * @author   Anant Narayanan <anant@php.net>
 * @license  http://opensource.org/licenses/BSD-3-Clause 3-clause BSD
 * @link     https://github.com/firebase/php-jwt
 */

//  $this->load->library(['Key', 'jwk']);
// namespace Libraries;
// use stdClass;
// use Libraries\Key;
// Use Firebase\JWT\Key;
// use ArrayAccess;

class Jwt
{
    /**
     * When checking nbf, iat or expiration times,
     * we want to provide some extra leeway time to
     * account for clock skew.
     */
    private static $asn1Integer = 0x02;
    private static $asn1Sequence = 0x10;
    private static $asn1BitString = 0x03;
    public static $leeway = 2000;
    /**
     * Allow the current timestamp to be specified.
     * Useful for fixing a value within unit testing.
     *
     * Will default to PHP time() value if null.
     */
    public static $timestamp = null;
    public static $supported_algs = array(
        'HS256' => array('hash_hmac', 'SHA256'),
        'HS512' => array('hash_hmac', 'SHA512'),
        'HS384' => array('hash_hmac', 'SHA384'),
        'RS256' => array('openssl', 'SHA256'),
        'RS384' => array('openssl', 'SHA384'),
        'RS512' => array('openssl', 'SHA512'),
    );
    /**
     * Decodes a JWT string into a PHP object.
     *
     * @param string        $jwt            The JWT
     * @param string|array  $key            The key, or map of keys.
     *                                      If the algorithm used is asymmetric, this is the public key
     * @param array         $allowed_algs   List of supported verification algorithms
     *                                      Supported algorithms are 'HS256', 'HS384', 'HS512' and 'RS256'
     *
     * @return object The JWT's payload as a PHP object
     *
     *
     * @uses jsonDecode
     * @uses urlsafeB64Decode
     */
    // public static function decode($jwt, $key, array $allowed_algs = array())
    // {
    //     $timestamp = is_null(static::$timestamp) ? time() : static::$timestamp;
    //     if (empty($key)) {
    //         throw new Exception('Key may not be empty');
    //     }
    //     $tks = explode('.', $jwt);
    //     if (count($tks) != 3) {
    //         throw new Exception('Wrong number of segments');
    //     }
    //     list($headb64, $bodyb64, $cryptob64) = $tks;
    //     if (null === ($header = static::jsonDecode(static::urlsafeB64Decode($headb64)))) {
    //         throw new Exception('Invalid header encoding');
    //     }
    //     if (null === $payload = static::jsonDecode(static::urlsafeB64Decode($bodyb64))) {
    //         throw new Exception('Invalid claims encoding');
    //     }
    //     if (false === ($sig = static::urlsafeB64Decode($cryptob64))) {
    //         throw new Exception('Invalid signature encoding');
    //     }
    //     if (empty($header->alg)) {
    //         throw new Exception('Empty algorithm');
    //     }
    //     if (empty(static::$supported_algs[$header->alg])) {
    //         throw new Exception('Algorithm not supported');
    //     }
    //     if (!in_array($header->alg, $allowed_algs)) {
    //         throw new Exception('Algorithm not allowed');
    //     }
    //     if (is_array($key) || $key instanceof \ArrayAccess) {
    //         if (isset($header->kid)) {
    //             if (!isset($key[$header->kid])) {
    //                 throw new Exception('"kid" invalid, unable to lookup correct key');
    //             }
    //             $key = $key[$header->kid];
    //         } else {
    //             throw new Exception('"kid" empty, unable to lookup correct key');
    //         }
    //     }
    //     // Check the signature
    //     if (!static::verify("$headb64.$bodyb64", $sig, $key, $header->alg)) {
    //         throw new Exception('Signature verification failed');
    //     }
    //     // Check if the nbf if it is defined. This is the time that the
    //     // token can actually be used. If it's not yet that time, abort.
    //     if (isset($payload->nbf) && $payload->nbf > ($timestamp + static::$leeway)) {
    //         throw new Exception(
    //             'Cannot handle token prior to ' . date(DateTime::ISO8601, $payload->nbf)
    //         );
    //     }
    //     // Check that this token has been created before 'now'. This prevents
    //     // using tokens that have been created for later use (and haven't
    //     // correctly used the nbf claim).
    //     if (isset($payload->iat) && $payload->iat > ($timestamp + static::$leeway)) {
    //         throw new Exception(
    //             'Cannot handle token prior to ' . date(DateTime::ISO8601, $payload->iat)
    //         );
    //     }
    //     // Check if this token has expired.
    //     if (isset($payload->exp) && ($timestamp - static::$leeway) >= $payload->exp) {
    //         //throw new Exception('Expired token');   
    //         return false;
    //     }
    //     return $payload;
    // }
    public static function decode(
        string $jwt,
        $keyOrKeyArray
    ): stdClass {
        // Validate JWT
        $timestamp = \is_null(static::$timestamp) ? \time() : static::$timestamp;

        if (empty($keyOrKeyArray)) {
            throw new Exception('Key may not be empty');
        }
        $tks = \explode('.', $jwt);
        if (\count($tks) !== 3) {
            throw new Exception('Wrong number of segments');
        }
        list($headb64, $bodyb64, $cryptob64) = $tks;
        $headerRaw = static::urlsafeB64Decode($headb64);
        if (null === ($header = static::jsonDecode($headerRaw))) {
            throw new Exception('Invalid header encoding');
        }
        $payloadRaw = static::urlsafeB64Decode($bodyb64);
        if (null === ($payload = static::jsonDecode($payloadRaw))) {
            throw new Exception('Invalid claims encoding');
        }
        if (\is_array($payload)) {
            // prevent PHP Fatal Error in edge-cases when payload is empty array
            $payload = (object) $payload;
        }
        if (!$payload instanceof stdClass) {
            throw new Exception('Payload must be a JSON object');
        }
        $sig = static::urlsafeB64Decode($cryptob64);
        if (empty($header->alg)) {
            throw new Exception('Empty algorithm');
        }
        if (empty(static::$supported_algs[$header->alg])) {
            throw new Exception('Algorithm not supported');
        }

        $key = self::getKey($keyOrKeyArray, property_exists($header, 'kid') ? $header->kid : null);

        // Check the algorithm
        if (!self::constantTimeEquals($key->getAlgorithm(), $header->alg)) {
            // See issue #351
            throw new Exception('Incorrect key for this algorithm');
        }
        if (\in_array($header->alg, ['ES256', 'ES256K', 'ES384'], true)) {
            // OpenSSL expects an ASN.1 DER sequence for ES256/ES256K/ES384 signatures
            $sig = self::signatureToDER($sig);
        }
        if (!self::verify("{$headb64}.{$bodyb64}", $sig, $key->getKeyMaterial(), $header->alg)) {
            throw new Exception('Signature verification failed');
        }

        // Check the nbf if it is defined. This is the time that the
        // token can actually be used. If it's not yet that time, abort.
        if (isset($payload->nbf) && $payload->nbf > ($timestamp + static::$leeway)) {
            throw new Exception(
                'Cannot handle token prior to ' . \date(DateTime::ISO8601, $payload->nbf)
            );
        }

        // Check that this token has been created before 'now'. This prevents
        // using tokens that have been created for later use (and haven't
        // correctly used the nbf claim).
        if (isset($payload->iat) && $payload->iat > ($timestamp + static::$leeway)) {
            throw new Exception(
                'Cannot handle token prior to ' . \date(DateTime::ISO8601, $payload->iat)
            );
        }

        // Check if this token has expired.
        if (isset($payload->exp) && ($timestamp - static::$leeway) >= $payload->exp) {
            throw new Exception('Expired token');
        }

        return $payload;
    }
    private static function getKey($keyOrKeyArray, $kid = null)
    {
       
        if ($keyOrKeyArray instanceof Key) {
            return $keyOrKeyArray;
        }

        if (is_array($keyOrKeyArray) || $keyOrKeyArray instanceof ArrayAccess) {
            foreach ($keyOrKeyArray as $keyId => $key) {
                if (!$key instanceof Key) {
                    throw new Exception(
                        '$keyOrKeyArray must be an instance of Firebase\JWT\Key key or an '
                        . 'array of Firebase\JWT\Key keys'
                    );
                }
            }
            if (!isset($kid)) {
                throw new Exception('"kid" empty, unable to lookup correct key');
            }
            if (!isset($keyOrKeyArray[$kid])) {
                throw new Exception('"kid" invalid, unable to lookup correct key');
            }

            return $keyOrKeyArray[$kid];
        }

        throw new Exception(
            '$keyOrKeyArray must be an instance of Firebase\JWT\Key key or an '
            . 'array of Firebase\JWT\Key keys'
        );
    }
    private static function signatureToDER($sig)
    {
        // Separate the signature into r-value and s-value
        list($r, $s) = \str_split($sig, (int) (\strlen($sig) / 2));

        // Trim leading zeros
        $r = \ltrim($r, "\x00");
        $s = \ltrim($s, "\x00");

        // Convert r-value and s-value from unsigned big-endian integers to
        // signed two's complement
        if (\ord($r[0]) > 0x7f) {
            $r = "\x00" . $r;
        }
        if (\ord($s[0]) > 0x7f) {
            $s = "\x00" . $s;
        }

        return self::encodeDER(
            self::$asn1Sequence,
            self::encodeDER(self::$asn1Integer, $r) .
            self::encodeDER(self::$asn1Integer, $s)
        );
    }
    private static function encodeDER($type, $value)
    {
        $tag_header = 0;
        if ($type === self::$asn1Sequence) {
            $tag_header |= 0x20;
        }

        // Type
        $der = \chr($tag_header | $type);

        // Length
        $der .= \chr(\strlen($value));

        return $der . $value;
    }
    public static function constantTimeEquals($left, $right)
    {
        if (\function_exists('hash_equals')) {
            return \hash_equals($left, $right);
        }
        $len = \min(static::safeStrlen($left), static::safeStrlen($right));

        $status = 0;
        for ($i = 0; $i < $len; $i++) {
            $status |= (\ord($left[$i]) ^ \ord($right[$i]));
        }
        $status |= (static::safeStrlen($left) ^ static::safeStrlen($right));

        return ($status === 0);
    }

    /**
     * Converts and signs a PHP object or array into a JWT string.
     *
     * @param object|array  $payload    PHP object or array
     * @param string        $key        The secret key.
     *                                  If the algorithm used is asymmetric, this is the private key
     * @param string        $alg        The signing algorithm.
     *                                  Supported algorithms are 'HS256', 'HS384', 'HS512' and 'RS256'
     * @param mixed         $keyId
     * @param array         $head       An array with header elements to attach
     *
     * @return string A signed JWT
     *
     * @uses jsonEncode
     * @uses urlsafeB64Encode
     */
    // public static function encode($payload, $key, $alg = 'HS256', $keyId = null, $head = null)
    // {
    //     $header = array('typ' => 'JWT', 'alg' => $alg);
    //     if ($keyId !== null) {
    //         $header['kid'] = $keyId;
    //     }
    //     if ( isset($head) && is_array($head) ) {
    //         $header = array_merge($head, $header);
    //     }
    //     $segments = array();
    //     $segments[] = static::urlsafeB64Encode(static::jsonEncode($header));
    //     $segments[] = static::urlsafeB64Encode(static::jsonEncode($payload));
    //     $signing_input = implode('.', $segments);
    //     $signature = static::sign($signing_input, $key, $alg);
    //     $segments[] = static::urlsafeB64Encode($signature);
    //     return implode('.', $segments);
    // }
    public static function encode(
        array $payload,
        $key,
        string $alg = 'HS256',
        string $keyId = null,
        array $head = null
    ): string {
        $header = ['typ' => 'JWT', 'alg' => $alg];
        if ($keyId !== null) {
            $header['kid'] = $keyId;
        }
        if (isset($head) && \is_array($head)) {
            $header = \array_merge($head, $header);
        }
        $segments = [];
        $segments[] = static::urlsafeB64Encode((string) static::jsonEncode($header));
        $segments[] = static::urlsafeB64Encode((string) static::jsonEncode($payload));
        $signing_input = \implode('.', $segments);

        $signature = static::sign($signing_input, $key, $alg);
        $segments[] = static::urlsafeB64Encode($signature);

        return \implode('.', $segments);
    }

    /**
     * Sign a string with a given key and algorithm.
     *
     * @param string            $msg    The message to sign
     * @param string|resource   $key    The secret key
     * @param string            $alg    The signing algorithm.
     *                                  Supported algorithms are 'HS256', 'HS384', 'HS512' and 'RS256'
     *
     * @return string An encrypted message
     *
     * @throws DomainException Unsupported algorithm was specified
     */
    public static function sign($msg, $key, $alg = 'HS256')
    {
        if (empty(static::$supported_algs[$alg])) {
            throw new DomainException('Algorithm not supported');
        }
        list($function, $algorithm) = static::$supported_algs[$alg];
        switch($function) {
            case 'hash_hmac':
                return hash_hmac($algorithm, $msg, $key, true);
            case 'openssl':
                $signature = '';
                $success = openssl_sign($msg, $signature, $key, $algorithm);
                if (!$success) {
                    throw new DomainException("OpenSSL unable to sign data");
                } else {
                    return $signature;
                }
        }
    }
    /**
     * Verify a signature with the message, key and method. Not all methods
     * are symmetric, so we must have a separate verify and sign method.
     *
     * @param string            $msg        The original message (header and body)
     * @param string            $signature  The original signature
     * @param string|resource   $key        For HS*, a string key works. for RS*, must be a resource of an openssl public key
     * @param string            $alg        The algorithm
     *
     * @return bool
     *
     * @throws DomainException Invalid Algorithm or OpenSSL failure
     */

     //php-JWt v-5 function
    // private static function verify($msg, $signature, $key, $alg)
    // {
    //     if (empty(static::$supported_algs[$alg])) {
    //         throw new DomainException('Algorithm not supported');
    //     }
    //     list($function, $algorithm) = static::$supported_algs[$alg];
    //     switch($function) {
    //         case 'openssl':
    //             $success = openssl_verify($msg, $signature, $key, $algorithm);
    //             if ($success === 1) {
    //                 return true;
    //             } elseif ($success === 0) {
    //                 return false;
    //             }
    //             // returns 1 on success, 0 on failure, -1 on error.
    //             throw new DomainException(
    //                 'OpenSSL error: ' . openssl_error_string()
    //             );
    //         case 'hash_hmac':
    //         default:
    //             $hash = hash_hmac($algorithm, $msg, $key, true);
    //             if (function_exists('hash_equals')) {
    //                 return hash_equals($signature, $hash);
    //             }
    //             $len = min(static::safeStrlen($signature), static::safeStrlen($hash));
    //             $status = 0;
    //             for ($i = 0; $i < $len; $i++) {
    //                 $status |= (ord($signature[$i]) ^ ord($hash[$i]));
    //             }
    //             $status |= (static::safeStrlen($signature) ^ static::safeStrlen($hash));
    //             return ($status === 0);
    //     }
    // }

    private static function verify($msg, $signature, $key, $alg)
    {
        if (empty(static::$supported_algs[$alg])) {
            throw new DomainException('Algorithm not supported');
        }

        list($function, $algorithm) = static::$supported_algs[$alg];
        switch ($function) {
            case 'openssl':
                $success = \openssl_verify($msg, $signature, $key, $algorithm);
                if ($success === 1) {
                    return true;
                } elseif ($success === 0) {
                    return false;
                }
                // returns 1 on success, 0 on failure, -1 on error.
                throw new DomainException(
                    'OpenSSL error: ' . \openssl_error_string()
                );
            case 'sodium_crypto':
              if (!function_exists('sodium_crypto_sign_verify_detached')) {
                  throw new DomainException('libsodium is not available');
              }
              try {
                  // The last non-empty line is used as the key.
                  $lines = array_filter(explode("\n", $key));
                  $key = base64_decode(end($lines));
                  return sodium_crypto_sign_verify_detached($signature, $msg, $key);
              } catch (Exception $e) {
                  throw new DomainException($e->getMessage(), 0, $e);
              }
            case 'hash_hmac':
            default:
                $hash = \hash_hmac($algorithm, $msg, $key, true);
                return self::constantTimeEquals($signature, $hash);
        }
    }
    /**
     * Decode a JSON string into a PHP object.
     *
     * @param string $input JSON string
     *
     * @return object Object representation of JSON string
     *
     * @throws DomainException Provided string was invalid JSON
     */

    // public static function jsonDecode($input)
    // {
    //     if (version_compare(PHP_VERSION, '5.4.0', '>=') && !(defined('JSON_C_VERSION') && PHP_INT_SIZE > 4)) {
    //         /** In PHP >=5.4.0, json_decode() accepts an options parameter, that allows you
    //          * to specify that large ints (like Steam Transaction IDs) should be treated as
    //          * strings, rather than the PHP default behaviour of converting them to floats.
    //          */
    //         $obj = json_decode($input, false, 512, JSON_BIGINT_AS_STRING);
    //     } else {
    //         /** Not all servers will support that, however, so for older versions we must
    //          * manually detect large ints in the JSON string and quote them (thus converting
    //          *them to strings) before decoding, hence the preg_replace() call.
    //          */
    //         $max_int_length = strlen((string) PHP_INT_MAX) - 1;
    //         $json_without_bigints = preg_replace('/:\s*(-?\d{'.$max_int_length.',})/', ': "$1"', $input);
    //         $obj = json_decode($json_without_bigints);
    //     }
    //     if (function_exists('json_last_error') && $errno = json_last_error()) {
    //         static::handleJsonError($errno);
    //     } elseif ($obj === null && $input !== 'null') {
    //         throw new DomainException('Null result with non-null input');
    //     }
    //     return $obj;
    // }

    public static function jsonDecode($input)
    {
        if (\version_compare(PHP_VERSION, '5.4.0', '>=') && !(\defined('JSON_C_VERSION') && PHP_INT_SIZE > 4)) {
            /** In PHP >=5.4.0, json_decode() accepts an options parameter, that allows you
             * to specify that large ints (like Steam Transaction IDs) should be treated as
             * strings, rather than the PHP default behaviour of converting them to floats.
             */
            $obj = \json_decode($input, false, 512, JSON_BIGINT_AS_STRING);
        } else {
            /** Not all servers will support that, however, so for older versions we must
             * manually detect large ints in the JSON string and quote them (thus converting
             *them to strings) before decoding, hence the preg_replace() call.
             */
            $max_int_length = \strlen((string) PHP_INT_MAX) - 1;
            $json_without_bigints = \preg_replace('/:\s*(-?\d{'.$max_int_length.',})/', ': "$1"', $input);
            $obj = \json_decode($json_without_bigints);
        }

        if ($errno = \json_last_error()) {
            static::handleJsonError($errno);
        } elseif ($obj === null && $input !== 'null') {
            throw new DomainException('Null result with non-null input');
        }
        return $obj;
    }
    /**
     * Encode a PHP object into a JSON string.
     *
     * @param object|array $input A PHP object or array
     *
     * @return string JSON representation of the PHP object or array
     *
     * @throws DomainException Provided object could not be encoded to valid JSON
     */

    // public static function jsonEncode($input)
    // {
    //     $json = json_encode($input);
    //     if (function_exists('json_last_error') && $errno = json_last_error()) {
    //         static::handleJsonError($errno);
    //     } elseif ($json === 'null' && $input !== null) {
    //         throw new DomainException('Null result with non-null input');
    //     }
    //     return $json;
    // }

    public static function jsonEncode($input)
    {
        if (PHP_VERSION_ID >= 50400) {
            $json = \json_encode($input, \JSON_UNESCAPED_SLASHES);
        } else {
            // PHP 5.3 only
            $json = \json_encode($input);
        }
        if ($errno = \json_last_error()) {
            static::handleJsonError($errno);
        } elseif ($json === 'null' && $input !== null) {
            throw new DomainException('Null result with non-null input');
        }
        return $json;
    }
    /**
     * Decode a string with URL-safe Base64.
     *
     * @param string $input A Base64 encoded string
     *
     * @return string A decoded string
     */
    public static function urlsafeB64Decode($input)
    {
        $remainder = strlen($input) % 4;
        if ($remainder) {
            $padlen = 4 - $remainder;
            $input .= str_repeat('=', $padlen);
        }
        return base64_decode(strtr($input, '-_', '+/'));
    }
    /**
     * Encode a string with URL-safe Base64.
     *
     * @param string $input The string you want encoded
     *
     * @return string The base64 encode of what you passed in
     */
    public static function urlsafeB64Encode($input)
    {
        return str_replace('=', '', strtr(base64_encode($input), '+/', '-_'));
    }
    /**
     * Helper method to create a JSON error.
     *
     * @param int $errno An error number from json_last_error()
     *
     * @return void
     */
    private static function handleJsonError($errno)
    {
        $messages = array(
            JSON_ERROR_DEPTH => 'Maximum stack depth exceeded',
            JSON_ERROR_STATE_MISMATCH => 'Invalid or malformed JSON',
            JSON_ERROR_CTRL_CHAR => 'Unexpected control character found',
            JSON_ERROR_SYNTAX => 'Syntax error, malformed JSON',
            JSON_ERROR_UTF8 => 'Malformed UTF-8 characters' //PHP >= 5.3.3
        );
        throw new DomainException(
            isset($messages[$errno])
            ? $messages[$errno]
            : 'Unknown JSON error: ' . $errno
        );
    }
    /**
     * Get the number of bytes in cryptographic strings.
     *
     * @param string
     *
     * @return int
     */
    private static function safeStrlen($str)
    {
        if (function_exists('mb_strlen')) {
            return mb_strlen($str, '8bit');
        }
        return strlen($str);
    }

    /**
        * Get hearder Authorization
        * */
        public function getAuthorizationHeader(){
            $headers = null;
            if (isset($_SERVER['Authorization'])) {
                $headers = trim($_SERVER["Authorization"]);
            }
            else if (isset($_SERVER['HTTP_AUTHORIZATION'])) { //Nginx or fast CGI
                $headers = trim($_SERVER["HTTP_AUTHORIZATION"]);
            } elseif (function_exists('apache_request_headers')) {
                $requestHeaders = apache_request_headers();
                // Server-side fix for bug in old Android versions (a nice side-effect of this fix means we don't care about capitalization for Authorization)
                $requestHeaders = array_combine(array_map('ucwords', array_keys($requestHeaders)), array_values($requestHeaders));
                if (isset($requestHeaders['Authorization'])) {
                    $headers = trim($requestHeaders['Authorization']);
                }
            }
            return $headers;
        }
        /**
         * get access token from header
         * */
        public function getBearerToken() {
            $headers = $this->getAuthorizationHeader();
            // HEADER: Get the access token from the header
            if (!empty($headers)) {
                if (preg_match('/Bearer\s(\S+)/', $headers, $matches)) {
                    return $matches[1];
                }
            }
            throw new DomainException('Invalid Hash or Hash not found');
        }
}
